package tantra;

import java.util.*;
import java.text.*;

/**
 * 日付データの保持や操作を行います。
 * 
 * @author Luft
 * 
 */
public class DateEx {

	/** このオブジェクトの本体 */
	private GregorianCalendar m_cal;

	/**
	 * <code>DateEx.toString</code>メソッドにおいて、出力される文字列のフォーマットを指定します。
	 * @author Luft
	 */
	public enum Format {
		/** 20071109のような形式の日付 */
		yyyyMMdd, 
		/** 21:59:21のような形式の時刻 */
		hh_mm_ss,
		/** 215921のような形式の時刻 */
		hhmmss,
		/** 2007/11/09のような形式の日付 */
		yyyy_MM_dd, 
		/** 2007/11/09 21:59:21のような形式の日時 */
		yyyy_MM_dd_hh_mm_ss
	}


	/**
	 * 出力される文字列の形式を指定し、日時を文字列に変換します。
	 * @param fmt 文字列の形式を指定する<code>enum</code>
	 * @return
	 * 		指定された形式でフォーマットされた日時を表した文字列
	 */
	public String toString( Format fmt ){
		DecimalFormat fmt4 = new DecimalFormat( "0000" );
		DecimalFormat fmt2 = new DecimalFormat( "00" );
		switch( fmt ){
			case yyyyMMdd:
				return fmt4.format( year() ) + fmt2.format( month() )
						+ fmt2.format( day() );
			case hh_mm_ss:
				return fmt2.format( hour() ) + ":" + fmt2.format( minute() )
						+ ":" + fmt2.format( second() );
			case hhmmss:
				return fmt2.format( hour() ) + fmt2.format( minute() )
						+ fmt2.format( second() );
			case yyyy_MM_dd:
				return fmt4.format( year() ) + "/" + fmt2.format( month() )
						+ "/" + fmt2.format( day() );
			case yyyy_MM_dd_hh_mm_ss:
				return fmt4.format( year() ) + "/" + fmt2.format( month() )
						+ "/" + fmt2.format( day() ) + " "
						+ fmt2.format( hour() ) + ":" + fmt2.format( minute() )
						+ ":" + fmt2.format( second() );
		}
		return "";
	}

	
	/**
	 * 日時を、2007/11/09 21:59:21のような形式で文字列に変換します
	 */
	public String toString(){
		DecimalFormat fmt4 = new DecimalFormat( "0000" );
		DecimalFormat fmt2 = new DecimalFormat( "00" );
		return fmt4.format( year() ) + "/" + fmt2.format( month() )
			+ "/" + fmt2.format( day() ) + " "
			+ fmt2.format( hour() ) + ":" + fmt2.format( minute() )
			+ ":" + fmt2.format( second() );
	}

	/**
	 * 値を指定したコンストラクタ．
	 * 
	 * @param year
	 *        年
	 * @param month
	 *        月
	 * @param day
	 *        日
	 * @param hour
	 *        時
	 * @param minute
	 *        分
	 * @param second
	 *        秒
	 */
	public DateEx( int year, int month, int day, int hour, int minute,
			int second ) {
		m_cal = new GregorianCalendar( year, month - 1, day, hour, minute,
				second );
	}


	/**
	 * 値を指定しないコンストラクタ。
	 * 西暦1年1月1日午前0時0分0秒を表す<code>DateEx</code>のインスタンスを生成します。
	 */
	public DateEx() {
		m_cal = setCalendar( 1, 1, 1, 0, 0, 0 );
	}


	/**
	 * <code>GregorianCalendar</code>に指定した日時をセットします。
	 * 日付が正規のものであるかどうかの判定は行われません。
	 * @param year 年
	 * @param month 月
	 * @param day 日
	 * @param hour 時
	 * @param minute 分
	 * @param second 秒
	 * @return セットされた日時。
	 */
	private static GregorianCalendar setCalendar( int year, int month, int day,
			int hour, int minute, int second ){
		return new GregorianCalendar( year, month - 1, day, hour, minute,
				second );
	}


	/**
	 * 文字列から時刻をコンストラクトします。文字列はyyyyMMddhhmmssの形式である必要があります。ex.20071103112554
	 * 文字列が必要な長さで無い場合や、日時として無効な文字列であった場合はDATE_NULLを返します。
	 * 
	 * @param iString
	 *        yyyyMMddhhmmss形式の文字列
	 * @return 取得した日時
	 */
	public DateEx( String iString ) {
		if( iString.length() < 14 ){
			// return new DateEx();
			m_cal = setCalendar( 1, 1, 1, 0, 0, 0 );
		}
		int y, m, d, hour, min, sec;
		y = Integer.valueOf( iString.substring( 0, 4 ) );
		m = Integer.valueOf( iString.substring( 4, 6 ) );
		d = Integer.valueOf( iString.substring( 6, 8 ) );
		hour = Integer.valueOf( iString.substring( 8, 10 ) );
		min = Integer.valueOf( iString.substring( 10, 12 ) );
		sec = Integer.valueOf( iString.substring( 12, 14 ) );
		if( isValidDate( y, m, d, hour, min, sec ) ){
			// return new DateEx( y, m, d, hour, min, sec );
			m_cal = setCalendar( y, m, d, hour, min, sec );
		}else{
			// return new DateEx();
			m_cal = setCalendar( 1, 1, 1, 0, 0, 0 );
		}
	}


	/**
	 * この日付に現在の日時をセットします。
	 */
	public void now(){
		m_cal = null;
		m_cal = new GregorianCalendar();
	}


	/**
	 * コピーを作成
	 */
	public DateEx clone(){
		return new DateEx( year(), month(), day(), hour(), minute(), second() );
	}


	/**
	 * この日付から時刻dを引き算します
	 * 
	 * @param d
	 *        この時刻から引き算する時刻。
	 * @return 秒数
	 */
	public long subtract( DateEx d ){
		return (m_cal.getTimeInMillis() - d.m_cal.getTimeInMillis()) / 1000;
	}


	/**
	 * この日付がnullであるかどうかを判定します。
	 * 
	 * @return この日付がnullであればtrueを、そうでなければfalseを返します。
	 */
	public boolean isNull(){
		if( year() == 1 && month() == 1 && day() == 1 ){
			return true;
		}else{
			return false;
		}
	}


	/**
	 * この日付が、日付<code>d</code>より新しいかどうかを調べます。
	 * 
	 * @param d
	 *        比較対象の日付。
	 * @return <code>d</code>よりも新しいものであればtrueを、そうでなければfalseを返します。
	 */
	public boolean largerThan( DateEx d ){
		int ymd1, ymd2;
		ymd1 = ymd();
		ymd2 = d.ymd();

		if( ymd1 > ymd2 ){
			return true;
		}else if( ymd1 == ymd2 ){
			ymd1 = hour() * 10000 + minute() * 100 + second();
			ymd2 = d.hour() * 10000 + d.minute() * 100 + d.second();
			if( ymd1 > ymd2 ){
				return true;
			}else{
				return false;
			}
		}else{
			return false;
		}
	}


	/**
	 * この日付が、日付<code>d</code>より新しいかどうかを調べます。
	 * 
	 * @param d
	 *        比較対象の日付。
	 * @return この日付が、日付<code>d</code>より新しいか等しければtrueを、そうでなければfalseを返します。
	 */
	public boolean largerOrEqual( DateEx d ){
		if( equals( d ) || largerThan( d ) ){
			return true;
		}
		return false;
	}


	/**
	 * この日付が、日付<code>d</code>と等しいかどうかを調べます。
	 * 
	 * @param d
	 *        比較対象の日付。
	 * @return この日付が、日付<code>d</code>と等しければtrueを、そうでなければfalseを返します。
	 */
	public boolean equals( DateEx d ){
		if( year() == d.year() && month() == d.month() && day() == d.day()
				&& hour() == d.hour() && minute() == d.minute()
				&& second() == d.second() ){
			return true;
		}
		return false;
	}


	/*
	 * "20070724"といった文字列から，DateEx型を得る．
	 * 
	 * @param token @return
	 */
	/*
	 * public static DateEx fromyyyyMMdd( String token ){ if( token.length() >=
	 * 8 ){ int i; char ch; for( i = 0; i <= 7; i++ ){ ch = token.charAt( i );
	 * if( !Character.isDigit( ch ) ) return new DateEx(); } int y, m, d; y =
	 * Integer.valueOf( token.substring( 0, 4 ) ); m = Integer.valueOf(
	 * token.substring( 4, 6 ) ); d = Integer.valueOf( token.substring( 6, 8 ) );
	 * if( isValidDate( y, m, d, 0, 0, 0 ) ){ return new DateEx( y, m, d, 0, 0,
	 * 0 ); }else{ return new DateEx(); } }else{ return new DateEx(); } }
	 */

	/**
	 * この日付に指定した日数を足し算した日付を返します。負の日数を指定することも可能です。
	 * 
	 * @param days
	 *        加算する日数
	 * @return 加算後の日付
	 */
	public DateEx addDays( int days ){
		DateEx tmpd = new DateEx( year(), month(), day(), hour(), minute(),
				second() );
		tmpd.m_cal.add( Calendar.DATE, days );
		return new DateEx( tmpd.year(), tmpd.month(), tmpd.day(), tmpd.hour(),
				tmpd.minute(), tmpd.second() );
	}


	/**
	 * この日付に指定した分を足し算した日付を返します。負の分を指定することも可能です。
	 * 
	 * @param minutes
	 *        加算する分
	 * @return 加算後の日時
	 */
	public DateEx addMinutes( int minutes ){
		DateEx tmpd = new DateEx( year(), month(), day(), hour(), minute(),
				second() );
		tmpd.m_cal.add( Calendar.MINUTE, minutes );
		return new DateEx( tmpd.year(), tmpd.month(), tmpd.day(), tmpd.hour(),
				tmpd.minute(), tmpd.second() );
	}


	/**
	 * edit1形式のログの行データから，その行が記録された年月日時刻を表すDate型変数を求めます
	 * 
	 * @param line
	 *        total data形式の行。ex."1,Luft1,Luft2,20070330,00:00:00"
	 * @return 取得した日時
	 */
	public static DateEx dateFromTotalData( String line ){
		try{
			String token[];
			if( line.length() <= 0 ){
				return new DateEx();
			}
			String edit = "";
			token = line.split( "," );
			edit = token[token.length - 2];
			int y, m, d;
			y = Integer.valueOf( edit.substring( 0, 4 ) );
			m = Integer.valueOf( edit.substring( 4, 6 ) );
			d = Integer.valueOf( edit.substring( 6, 8 ) );
			int hour, min, sec;
			edit = token[token.length - 1];
			String token2[];
			token2 = edit.split( ":" );
			hour = Integer.valueOf( token2[0] );
			min = Integer.valueOf( token2[1] );
			sec = Integer.valueOf( token2[2] );
			if( isValidDate( y, m, d, hour, min, sec ) ){
				return new DateEx( y, m, d, hour, min, sec );
			}else{
				return new DateEx();
			}
		}catch( Exception e ){
			tantra.printError( e, line );
			return new DateEx();
		}
	}


	/**
	 * この日付が含まれる週の週の初めの日付を返します。時刻は00:00:00にセットされます。
	 * 
	 * @return この日付が含まれる週の日曜日の日時。
	 */
	public DateEx weekStart(){
		if( this.isNull() ){
			return new DateEx();
		}
		DateEx tmpd = new DateEx( year(), month(), day(), 0, 0, 0 );
		int dow;
		dow = tmpd.dayOfWeek();
		tmpd = tmpd.addDays( -dow );
		/*
		 * System.out.println( "DateEx::getWeekStart;this,return=" +
		 * this.getyyyy_MM_dd() + " " + this.gethh_mm_ss() + "," +
		 * tmpd.getyyyy_MM_dd() + "," + tmpd.gethh_mm_ss() );
		 */
		return tmpd;
	}


	/**
	 * ChatLogの行末から当該行の記録された日時を取得します
	 * 
	 * @param line
	 *        ChatLogに記録された行。ex."[告知]Luft1 が Luft2 に勝利しました。 (20070506,23:37:54)"
	 * @return 取得した日付。ex.20070506,23:37:54
	 */
	public static DateEx dateFromChatLog( String line ){
		// [告知]Luft1 が Luft2 に勝利しました。 (20070506,23:37:54)
		try{
			if( line.length() < 19 ){
				return new DateEx();
			}
			String edit = line.substring( line.length() - 19 ); // (20070506,23:37:54)
			int year, day, month, hour, minute, second;
			year = Integer.valueOf( edit.substring( 1, 5 ) );
			month = Integer.valueOf( edit.substring( 5, 7 ) );
			day = Integer.valueOf( edit.substring( 7, 9 ) );
			hour = Integer.valueOf( edit.substring( 10, 12 ) );
			minute = Integer.valueOf( edit.substring( 13, 15 ) );
			second = Integer.valueOf( edit.substring( 16, 18 ) );
			if( isValidDate( year, month, day, hour, minute, second ) ){
				return new DateEx( year, month, day, hour, minute, second );
			}else{
				return new DateEx();
			}
		}catch( Exception e ){
			tantra.printError( e, line );
			return new DateEx();
		}
	}


	/**
	 * この日付の曜日を表すDayOfWeek値を求めます。日曜日が0、土曜日が6となります。
	 * 
	 * @return この日付の曜日を表すDayOfWeek値
	 */
	public int dayOfWeek(){
		return m_cal.get( Calendar.DAY_OF_WEEK ) - Calendar.SUNDAY;
	}


	/**
	 * この日付の年月日を20071101といった形式の整数に変換します
	 * 
	 * @return この日付を表す整数値
	 */
	public int ymd(){
		return year() * 10000 + month() * 100 + day();
	}


	/**
	 * この日付の年を返します。
	 * 
	 * @return この日付の年。
	 */
	public int year(){
		return m_cal.get( Calendar.YEAR );
	}


	/**
	 * この日付の月を返します。
	 * 
	 * @return この日付の月。
	 */
	public int month(){
		return m_cal.get( Calendar.MONTH ) + 1;
	}


	/**
	 * この日付の日を返します。
	 * 
	 * @return この日付の日。
	 */
	public int day(){
		return m_cal.get( Calendar.DATE );
	}


	/**
	 * この日付の時間を返します。（24時間制）
	 * 
	 * @return この日付の時間。
	 */
	public int hour(){
		return m_cal.get( Calendar.HOUR_OF_DAY );
	}


	/**
	 * この日付の分を返します。
	 * 
	 * @return この日付の分。
	 */
	public int minute(){
		return m_cal.get( Calendar.MINUTE );
	}


	/**
	 * この日付の秒を返します。
	 * 
	 * @return この日付の秒。
	 */
	public int second(){
		return m_cal.get( Calendar.SECOND );
	}


	/**
	 * 指定された日付が正当なものかどうかを判定します。
	 * 
	 * @param y
	 *        年
	 * @param m
	 *        月
	 * @param d
	 *        日
	 * @param hour
	 *        時
	 * @param min
	 *        分
	 * @param sec
	 *        秒
	 * @return 正当なものならtrueを、そうでなければfalseを返します。
	 */
	public static boolean isValidDate( int y, int m, int d, int hour, int min,
			int sec ){
		final int dmonth[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30,
				31 };
		if( m < 1 || 12 < m ) return false;
		if( m == 2 ){
			GregorianCalendar t_cal = new GregorianCalendar();
			if( t_cal.isLeapYear( y ) ){
				if( d < 1 || 29 < d ) return false;
			}else{
				if( d < 1 || 28 < d ) return false;
			}
		}else{
			if( d < 1 || dmonth[m] < d ) return false;
		}
		if( hour < 0 || 23 < hour ) return false;
		if( min < 0 || 59 < min ) return false;
		if( sec < 0 || 59 < sec ) return false;
		return true;
	}
}
