package tantra;

import java.util.*;
import java.io.*;
import tantra.DateEx.*;

/**
 * PersonalDataクラスは、キャラクタの名前、活動時間、勝ち・負け回数などを保持するためのもの。
 * 
 * @author Luft
 */
public class PersonalData {

	/** キャラクタ名 */
	public String name;
	/** 主神判定値。ブラフマ主神の可能性があればtrue */
	public boolean brahma;
	/** 主神判定値。ヴィシュヌ主神の可能性があればtrue */
	public boolean vishnu;
	/** 主神判定値。シヴァ主神の可能性があればtrue */
	public boolean shiva;
	/** 分で表した活動時間 */
	public double activityMin;
	/** このキャラクタが活動していることが確認された最新の時刻 */
	private DateEx m_dLastLog;
	/** 薬値順で判定された順位 */
	public int orderDoctor;
	/** DMG値と薬値の合計値順で判定された順位 */
	public int orderTotal;
	/** 種族 */
	//public Job job;
	/** 合計勝利回数 */
	public int win;
	/** 合計敗北回数 */
	public int lose;

	private final static double DMG_CONST = 0.05436267490685;
	private final static double DOCTOR_CONST = 0.07047043025644;


	/**
	 * コンストラクタ
	 * 
	 * @param string
	 *        キャラクタの名前。ex."Luft"
	 *        もしくは予備解析データから読み込んだ行データ。ex."Luft,1,0,0,0,1,10,20070330,00:00:20,0,0,Unknown "
	 */
	public PersonalData( String string ) {
		if( string.indexOf( "," ) < 0 ){
			/* 通常のコンストラクタの動作。キャラクタ名にはコンマが入らないことを利用 */
			name = string;
			brahma = true;
			vishnu = true;
			shiva = true;
			win = 0;
			lose = 0;
			activityMin = 0.0;
			m_dLastLog = new DateEx( 1, 1, 1, 0, 0, 0 );
			orderDoctor = 0;
			orderTotal = 0;
			//job = Job.Unknown;
		}else{
			/* 予備解析データ行を直接使ったコンストラクタ */
			// スマン、PersonalDataのbrahma, vishnu, shivaをintからbooleanにしたときに、
			// writeの方でそのまま書いちゃったからめんどくさいことに。
			String token[];
			token = string.split( "," );
			name = token[0];
			if( token[1].equals( "true" ) ){
				brahma = true;
			}else if( token[1].equals( "false" ) ){
				brahma = false;
			}else if( Integer.valueOf( token[1] ).intValue() == 1 ){
				brahma = true;
			}else{
				brahma = false;
			}
			if( token[2].equals( "true" ) ){
				brahma = true;
			}else if( token[2].equals( "false" ) ){
				brahma = false;
			}else if( Integer.valueOf( token[2] ).intValue() == 1 ){
				vishnu = true;
			}else{
				vishnu = false;
			}
			if( token[3].equals( "true" ) ){
				brahma = true;
			}else if( token[3].equals( "false" ) ){
				brahma = false;
			}else if( Integer.valueOf( token[3] ).intValue() == 1 ){
				shiva = true;
			}else{
				shiva = false;
			}
			win = Integer.valueOf( token[4] ).intValue();
			lose = Integer.valueOf( token[5] ).intValue();
			activityMin = Double.valueOf( token[6] ).intValue();
			m_dLastLog = DateEx.dateFromChatLog( "(" + token[7] + ","
					+ token[8] + ")" );
			orderDoctor = 0;
			orderTotal = 0;
			//job = Job.valueOf( token[11] );
		}
	}


	/**
	 * 主神判定値を合計した値を返します
	 * 
	 * @return 主神判定値の合計値
	 */
	public int sumsLordValue(){
		int result = 0;
		if( brahma ) result += 1;
		if( vishnu ) result += 1;
		if( shiva ) result += 1;
		return result;
	}


	/**
	 * PersonalDataが等しいかどうかを判定します
	 * 
	 * @param pd
	 *        比較するPersonalData
	 * @return pdとthisが等しければtrueを、そうでなければfalseを返します
	 */
	public boolean equals( PersonalData pd ){
		if( name.equals( pd.name ) ){
			return true;
		}else{
			return false;
		}
	}


	/**
	 * キルログの勝者・敗者の情報から，主神判別値を更新します
	 * 
	 * @param pd1
	 *        判定するキャラクタその1
	 * @param pd2
	 *        判定するキャラクタその2
	 */
	public static void manageLordValue( PersonalData pd1, PersonalData pd2 ){
		int slv1, slv2;
		slv1 = pd1.sumsLordValue();
		slv2 = pd2.sumsLordValue();
		if( slv1 == 1 && slv2 == 1 ){
			if( pd1.brahma == pd2.brahma && pd1.vishnu == pd2.vishnu
					&& pd1.shiva == pd2.shiva ){
				// 敵対しているのに同じ主神値を持っていた場合
				pd1.brahma = pd1.vishnu = pd1.shiva = true;
				pd2.brahma = pd2.vishnu = pd2.shiva = true;
			}
		}else if( slv1 == 1 && slv2 > 1 ){
			if( pd1.brahma ) pd2.brahma = false;
			if( pd1.vishnu ) pd2.vishnu = false;
			if( pd1.shiva ) pd2.shiva = false;
		}else if( slv1 > 1 && slv2 == 1 ){
			if( pd2.brahma ) pd1.brahma = false;
			if( pd2.vishnu ) pd1.vishnu = false;
			if( pd2.shiva ) pd1.shiva = false;
		}
	}


	/**
	 * PersonalDataの各主神値からそのプレイヤーの主神を判定します． 判定できない場合はLord.Unknownを返します
	 * 
	 * @return このキャラクタの主神値から判定した所属主神
	 */
	public Lord detectLord(){
		int slv = sumsLordValue();
		if( slv == 1 ){
			if( brahma ) return Lord.Brahma;
			if( vishnu ) return Lord.Vishnu;
			if( shiva ) return Lord.Shiva;
		}
		return Lord.Unknown;
	}


	/**
	 * キャラクタの活動時間を計算します。
	 * 
	 * @param log_date
	 *        ログの発生した時刻
	 */
	public void calculateActivityTime( DateEx log_date ){
		if( log_date.isNull() ){
			activityMin = 5.0;
		}else{
			long interval = log_date.subtract( m_dLastLog );
			double mins = (double) interval / 60.0;
			if( mins >= 0.0 ){
				if( mins > tantra.DAY_INTERVAL ){
					activityMin += tantra.DAY_INTERVAL / 2.0;
				}else{
					activityMin += mins;
				}
			}
		}
		m_dLastLog = log_date;
	}


	/**
	 * 勝率を計算します
	 * 
	 * @return このキャラクタの勝率を返します。勝敗のデータを保持していない場合は 0.0 を返します。
	 */
	private double winningPercentage(){
		if( win + lose > 0 ){
			return (double) win / ((double) win + (double) lose);
		}else{
			return 0.0;
		}
	}


	/**
	 * DMG値を計算します。
	 * 
	 * @return このキャラクタのDMG値を返します。
	 */
	public double dmg(){
		return Math.exp( Math.E * winningPercentage() ) * (win * Math.E + lose)
				* DMG_CONST;
	}


	/**
	 * 薬値を計算します。
	 * 
	 * @return このキャラクタの薬値を返します。
	 */
	public double doctor(){
		return activityMin * Math.pow( 1.0 - winningPercentage(), 7.0 )
				* DOCTOR_CONST;
	}


	/**
	 * 順位を判定する値を計算します。薬値を考慮した値を計算するかどうかを、modeで指定します。
	 * 
	 * @param mode
	 *        列挙型Modeの値。
	 * @return modeがDOCTORのときは薬値も加算した値を、そうでないときはDMG値のみを返します。
	 */
	public double ratingValue( tantra.OutputMode mode ){
		switch( mode ){
			case DOCTOR:
				return doctor();
			default:
				return doctor() + dmg();
		}
	}


	/**
	 * pdに格納されているPersonalDataを、ファイルに保存します。
	 * 
	 * @param pd
	 *        PersonalData型オブジェクトが格納されているVector
	 * @param fname
	 *        保存するファイルの名前
	 * @param last_line
	 */
	public static void write( Vector<PersonalData> pd, String fname,
			String last_line ){
		if( pd.size() >= 1 ){
			try{
				BufferedWriter sw = new BufferedWriter( new OutputStreamWriter(
						new FileOutputStream( fname ), tantra.ShiftJIS ) );
				sw.write( last_line );
				sw.newLine();

				int i;
				PersonalData tpd;
				for( i = 0; i < pd.size(); i++ ){
					tpd = pd.elementAt( i );
					sw.write( tpd.name + "," + tantra.intValue( tpd.brahma ) + "," + tantra.intValue( tpd.vishnu )
							+ "," + tantra.intValue( tpd.shiva ) + "," + tpd.win + "," + tpd.lose
							+ "," + tpd.activityMin + ","
							+ tpd.m_dLastLog.toString( Format.yyyyMMdd ) + ","
							+ tpd.m_dLastLog.toString( Format.hh_mm_ss ) + ","
							+ tpd.orderTotal + "," + tpd.orderDoctor + ","
							+ "Unknown" /*tpd.job.toString()*/ );
					sw.newLine();
				}
				sw.write( "END" );
				sw.newLine();

				sw.close();
			}catch( Exception e ){
				tantra.printError( e );
			}
		}
	}


	/**
	 * 配列に格納されているPersonalDataの中から、キャラクタ名が<code>person</code>であるものの
	 * 名前を検索します。当該キャラクタが見つからず、かつ<code>rewritable</code>がtrueに指定されている
	 * 場合は、pdに要素を追加します。
	 * 
	 * @param pd
	 *        <code>PersonalData</code>が格納されているVector
	 * @param person
	 *        キャラクタ名
	 * @param rewritable
	 *        <code>pd</code>の変更を許可する場合はtrueを、そうでない場合はfalseを指定します。
	 * @return pdの中にキャラクタが見つかった場合はそのインデックスを、見つからなかった場合は-1を返します。 ただし、<code>rewritable</code>がtrueの場合は要素を追加し、追加した要素のインデックスを返します。
	 */
	public static int indexOf( Vector<PersonalData> pd, String person,
			boolean rewritable ){
		PersonalData tpd;
		if( pd.size() <= 0 ){
			if( rewritable ){
				if( pd.add( new PersonalData( person ) ) ){
					return 0;
				}else{
					return -1;
				}
			}else{
				return -1;
			}
		}else{
			int i;
			int length = pd.size() - 1;
			for( i = 0; i <= length; i++ ){
				tpd = (PersonalData) pd.elementAt( i );
				if( tpd.name.equals( person ) ){
					return i;
				}
			}
			if( rewritable ){
				pd.add( new PersonalData( person ) );
				return length + 1;
			}else{
				return -1;
			}
		}
	}

}
