package tantra;

import java.io.*;

/**
 * UWSCが保存するログファイルから、一個ずつログ・エントリーを読み込む。
 * ロックファイルを使って同時にファイルを開かないように注意しながら、読み込んでいく。
 * 
 * @author Luft
 */
public class UwscLogServer {

	private RandomAccessFile UWSCLOG = null;
	private boolean locked = false;
	public long uwsc_current_record = 0;
	private boolean opened = false;
	private final int RECORD_LEN = 21;


	// private long uwsc_length;
	// private final static int uwsc_record_length = 21;

	/**
	 * uwscログファイルから次のエントリーを読み込みます
	 * 読込に成功した場合は読み込んだエントリーを，失敗した場合はUWSCLOGENTRY_NULLを返します
	 * 
	 * @return
	 */
	public UwscLogEntry read(){
		if( !opened ) return new UwscLogEntry();
		if( !locked ) enterLockMode();
//		String line;
		byte[] buffer = new byte[RECORD_LEN];
		//UwscLogEntry entry = new UwscLogEntry();
		try{
//			if( (line = UWSCLOG.readLine()) != null ){
			UWSCLOG.seek( uwsc_current_record * RECORD_LEN );
			int result = UWSCLOG.read( buffer );
			//System.out.println( "result=" + result );
			if( result == RECORD_LEN ){
				uwsc_current_record += 1;
				if( locked ) exitLockMode();
				String line = new String( buffer );
				//System.out.println( "read;line=" + line );
				return new UwscLogEntry( new String( buffer ) );
			}else{
				if( locked ) exitLockMode();
				return new UwscLogEntry();
			}
		}catch( Exception e ){
			tantra.printError( e );
			if( locked ) exitLockMode();
			return new UwscLogEntry();
		}
	}


	/**
	 * uwscログファイルをロックモードにします．
	 * 具体的には，"lockfile.bin"を生成してtantra_data.uwsによる書き込みを抑制します． 失敗した場合はfalseを返します
	 * 
	 * @return ロックできればtrueを、それ以外はfalseを返します
	 */
	public boolean enterLockMode(){
		String fpath_uwsc = tantra.settings.uwsc_location + File.separator
				+ tantra.LOCKFILE_UWSC;
		String fpath_exe = tantra.settings.uwsc_location + File.separator
				+ tantra.LOCKFILE_EXE;
		// System.out.println( "UwscLogServer::enterLockMode;" );
		if( tantra.exists( fpath_uwsc ) ){
			locked = false;
			return false;
		}
		File fp = new File( fpath_exe );
		try{
			if( fp.exists() ){
				locked = true;
				return true;
			}else{
				locked = fp.createNewFile();
				return locked;
			}
		}catch( Exception e ){
			tantra.printError( e );
			locked = false;
			return false;
		}
	}


	/**
	 * uwscログファイルのロックモードを解除します
	 * 
	 * @return 解除に成功すればtrueを、それ以外はfalseを返します
	 */
	public boolean exitLockMode(){
		File fp = new File( tantra.settings.uwsc_location + File.separator
				+ tantra.LOCKFILE_EXE );
		if( fp.exists() ){
			try{
				boolean result = fp.delete();
				if( result ) locked = false;
				return result;
			}catch( Exception e ){
				tantra.printError( e );
				locked = true;
				return false;
			}
		}
		locked = false;
		return false;
	}


	/**
	 * uwscログファイルを開きます
	 * 
	 * @return ファイルオープンに成功すればtrueを、それ以外はfalseを返します
	 */
	public boolean openLogFile(){
		//System.out.println( "openLogFile" );
		String fname = tantra.settings.uwsc_location + File.separator
				+ tantra.UWSC_LOG;
		File fi = new File( fname );
		if( !fi.exists() ){
			opened = false;
			return false;
		}
		// uwsc_length = fi.length();
		try{
			//UWSCLOG = new BufferedReader( new InputStreamReader(
			//		new FileInputStream( fname ), tantra.ShiftJIS ) );
			UWSCLOG = new RandomAccessFile( fname, "r" );
			//System.out.println( "openLogFile;open successed" );
			opened = true;
			return true;
		}catch( Exception e ){
			tantra.printError( e );
			opened = false;
			return false;
		}
	}


	/**
	 * uwscログファイルを閉じます。
	 */
	public void closeLogFile(){
		if( UWSCLOG != null ){
			try{
				UWSCLOG.close();
				opened = false;
			}catch( Exception e ){
				opened = true;
				tantra.printError( e );
			}
		}
		opened = false;
	}
}
